@@ -9,25 +9,26 @@ import json
 import logging
 import os
 import tempfile
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import Any, Callable, List, Optional, Tuple, Union
 
 from ._libtpm2_pytss import ffi, lib
-from .callbacks import Callback, CallbackType, get_callback, unlock_callback
 from .fapi_info import FapiInfo
 from .internal.utils import _chkrc, _check_bug_fixed, _get_dptr, _to_bytes_or_null
 from .TCTI import TCTI
 from .types import TPM2B_PUBLIC, TPM2B_PRIVATE
+from .constants import TSS2_RC, TPM2_ALG
+from .TSS2_Exception import TSS2_Exception
 
 logger = logging.getLogger(__name__)
 
-ffi_malloc = ffi.new_allocator(free=None)
-
 FAPI_CONFIG_ENV = "TSS2_FAPICONF"
 FAPI_CONFIG_PATHS = [
     "/etc/tpm2-tss/fapi-config.json",
     "/usr/local/etc/tpm2-tss/fapi-config.json",
 ]
 
+_cffi_malloc = ffi.new_allocator(free=None)
+
 
 class FAPIConfig(contextlib.ExitStack):
     """Context to create a temporary Fapi environment."""
@@ -42,8 +43,8 @@ class FAPIConfig(contextlib.ExitStack):
         and/or if `temp_dirs` is True.
 
         Args:
-            config (dict, optional): Fapi configuration to use instead of the installed `fapi-config.json`. Defaults to None.
-            temp_dirs (bool, optional): Create temporary keystore and log directories and set the respective config entries. Defaults to True.
+            config (dict): Fapi configuration to use instead of the installed `fapi-config.json`. Defaults to None.
+            temp_dirs (bool): Create temporary keystore and log directories and set the respective config entries. Defaults to True.
             **kwargs: Single configuration entries which override those in `config` or `fapi-config.json`.
         """
         super().__init__()
@@ -94,10 +95,10 @@ class FAPIConfig(contextlib.ExitStack):
 
             self.config = {**self.config, **temp_dir_config}
 
-        fapi_conf_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
-        self.config_tmp_path = fapi_conf_file.name
-        fapi_conf_file.write(json.dumps(self.config))
-        fapi_conf_file.close
+        with tempfile.NamedTemporaryFile(mode="w", delete=False) as fapi_conf_file:
+            self.config_tmp_path = fapi_conf_file.name
+            fapi_conf_file.write(json.dumps(self.config))
+
         logger.debug(
             f"Creating FAPIConfig: {self.config_tmp_path}:\n{json.dumps(self.config, indent=4)}"
         )
@@ -115,6 +116,121 @@ class FAPIConfig(contextlib.ExitStack):
             os.unlink(self.config_tmp_path)
 
 
+class _FAPI_CB_UDATA:
+    def __init__(self, cb, udata):
+        self.cur_exc = None
+        self.udata = udata
+        self.cb = cb
+
+
+@ffi.def_extern()
+def _fapi_auth_callback(object_path, description, auth, user_data):
+    cb_udata: _FAPI_CB_UDATA = ffi.from_handle(user_data)
+
+    if not cb_udata.cb:
+        return TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+    try:
+
+        got_auth: Optional[bytes, str] = cb_udata.cb(
+            ffi.string(object_path), ffi.string(description), cb_udata.udata
+        )
+
+        auth_bytes = got_auth.decode() if isinstance(got_auth, str) else got_auth
+        auth[0] = _cffi_malloc("char[]", auth_bytes)
+    except Exception as e:
+        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+        cb_udata.cur_exc = e
+        return rc
+
+    return TSS2_RC.RC_SUCCESS
+
+
+@ffi.def_extern()
+def _fapi_policy_action_callback(object_path, action, user_data):
+    cb_udata: _FAPI_CB_UDATA = ffi.from_handle(user_data)
+
+    if not cb_udata.cb:
+        return TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+    try:
+        cb_udata.cb(
+            ffi.string(object_path).decode(),
+            ffi.string(action).decode(),
+            cb_udata.udata,
+        )
+    except Exception as e:
+        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+        cb_udata.cur_exc = e
+        return rc
+
+    return TSS2_RC.RC_SUCCESS
+
+
+@ffi.def_extern()
+def _fapi_sign_callback(
+    object_path,
+    description,
+    publickey,
+    publickey_hint,
+    hashalg,
+    data_to_sign,
+    data_to_sign_size,
+    signature,
+    signature_size,
+    user_data,
+):
+    cb_udata: _FAPI_CB_UDATA = ffi.from_handle(user_data)
+
+    if not cb_udata.cb:
+        return TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+    try:
+        sig: bytes = cb_udata.cb(
+            ffi.string(object_path).decode(),
+            ffi.string(description).decode(),
+            ffi.string(publickey).decode(),
+            ffi.string(publickey_hint).decode(),
+            TPM2_ALG(hashalg),
+            ffi.buffer(data_to_sign, data_to_sign_size)[:],
+            cb_udata.udata,
+        )
+        signature_size[0] = len(sig)
+        signature[0] = _cffi_malloc("char[]", sig)
+    except Exception as e:
+        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.FAPI_RC_GENERAL_FAILURE
+        cb_udata.cur_exc = e
+        return rc
+
+    return TSS2_RC.RC_SUCCESS
+
+
+@ffi.def_extern()
+def _fapi_branch_callback(
+    object_path, description, branch_names, num_branches, selected_branch, user_data
+):
+    cb_udata: _FAPI_CB_UDATA = ffi.from_handle(user_data)
+
+    if not cb_udata.cb:
+        return TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+    try:
+
+        branch_list = [
+            ffi.string(x).decode() for x in ffi.unpack(branch_names, num_branches)
+        ]
+
+        position: int = cb_udata.cb(
+            ffi.string(object_path).decode(),
+            ffi.string(description).decode(),
+            branch_list,
+            cb_udata.udata,
+        )
+        selected_branch[0] = position
+    except Exception as e:
+        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.FAPI_RC_NOT_IMPLEMENTED
+        cb_udata.cur_exc = e
+        return rc
+
+    return TSS2_RC.RC_SUCCESS
+
+
 class FAPI:
     """The TPM2 Feature API. This class can be used as a python context or be closed manually via
     :meth:`~tpm2_pytss.FAPI.close`.
@@ -128,8 +244,7 @@ class FAPI:
         ret = lib.Fapi_Initialize(self._ctx_pp, uri)
         _chkrc(ret)
 
-        # set callbacks
-        self.callbacks: Dict[CallbackType, Optional[Callback]] = {}
+        self._callback_metadata = {}
 
     @property
     def _ctx(self):
@@ -150,11 +265,6 @@ class FAPI:
         """Finalize the Feature API. This frees allocated memory and invalidates the FAPI object."""
         lib.Fapi_Finalize(self._ctx_pp)
 
-        for callback_type, callback in self.callbacks.items():
-            if callback is not None:
-                unlock_callback(callback_type, callback.name)
-                self.callbacks[callback_type] = None
-
     # TODO flesh out info class
     @property
     def version(self):
@@ -191,10 +301,10 @@ class FAPI:
         objects. See also config file `/etc/tpm2-tss/fapi-config.json`.
 
         Args:
-            auth_value_eh (bytes or str, optional): Endorsement Hierarchy password. Defaults to None.
-            auth_value_sh (bytes or str, optional, optional): Storage/Owner Hierarchy password. Defaults to None.
-            auth_value_lockout (bytes or str, optional): Lockout Hierarchy password. Defaults to None.
-            is_provisioned_ok (bool, optional): Do not throw a TSS2_Exception if Fapi is already provisioned. Defaults to True.
+            auth_value_eh (bytes or str): Endorsement Hierarchy password. Defaults to None.
+            auth_value_sh (bytes or str): Storage/Owner Hierarchy password. Defaults to None.
+            auth_value_lockout (bytes or str): Lockout Hierarchy password. Defaults to None.
+            is_provisioned_ok (bool): Do not throw a TSS2_Exception if Fapi is already provisioned. Defaults to True.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -255,7 +365,7 @@ class FAPI:
         """Get a list of all Fapi current object paths.
 
         Args:
-            search_path (bytes or str, optional): If given, only list children of `search_path`. Defaults to None.
+            search_path (bytes or str): If given, only list children of `search_path`. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -285,9 +395,9 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the new key object, e.g. `/HS/SRK/new_signing_key`.
-            type_ (bytes or str, optional): Comma separated list. Possible values: system, sign, decrypt, restricted, exportable, noda, 0x81000000. Defaults to None.
-            auth_value (bytes or str, optional): Password to key. Defaults to None.
-            exists_ok (bool, optional): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
+            type_ (bytes or str): Comma separated list. Possible values: system, sign, decrypt, restricted, exportable, noda, 0x81000000. Defaults to None.
+            auth_value (bytes or str): Password to key. Defaults to None.
+            exists_ok (bool): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -316,7 +426,7 @@ class FAPI:
         Args:
             path (bytes or str): Path to the signing key.
             digest (bytes): Digest to sign.
-            padding (bytes or str, optional): `"rsa_ssa"` or `"rsa_pss"`. Defaults to None (using the scheme specified in the crypto profile).
+            padding (bytes or str): `"rsa_ssa"` or `"rsa_pss"`. Defaults to None (using the scheme specified in the crypto profile).
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -437,12 +547,12 @@ class FAPI:
 
         Args:
             path (bytes or str): The path of the new sealed object.
-            data (bytes or str, optional): Data to be sealed (often a digest). If None, random data will be generated. Defaults to None.
-            type_ (bytes or str, optional): Comma separated list. Possible values: system, sign, decrypt, restricted, exportable, noda, 0x81000000. Defaults to None.
-            policy_path (bytes or str, optional): The path to the policy which will be associated with the sealed object. Defaults to None.
-            auth_value (bytes or str, optional): Password to protect the new sealed object. Defaults to None.
-            size (int, optional): If data is None, random bytes of length size are generated. Parameters data and size cannot be given at the same time. Defaults to None.
-            exists_ok (bool, optional): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
+            data (bytes or str): Data to be sealed (often a digest). If None, random data will be generated. Defaults to None.
+            type_ (bytes or str): Comma separated list. Possible values: system, sign, decrypt, restricted, exportable, noda, 0x81000000. Defaults to None.
+            policy_path (bytes or str): The path to the policy which will be associated with the sealed object. Defaults to None.
+            auth_value (bytes or str): Password to protect the new sealed object. Defaults to None.
+            size (int): If data is None, random bytes of length size are generated. Parameters data and size cannot be given at the same time. Defaults to None.
+            exists_ok (bool): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -501,7 +611,7 @@ class FAPI:
         Args:
             path (bytes or str): Path of the future Fapi object.
             import_data (bytes or str): JSON-encoded data to import.
-            exists_ok (bool, optional): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
+            exists_ok (bool): Do not throw a TSS2_Exception if an object with the given path already exists. Defaults to False.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -541,7 +651,7 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the Fapi object.
-            auth_value (bytes or str, optional): New password. Defaults to None.
+            auth_value (bytes or str): New password. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -558,7 +668,7 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the existing Fapi object.
-            new_path (bytes or str, optional): New path to the Fapi object. Defaults to None.
+            new_path (bytes or str): New path to the Fapi object. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -580,7 +690,7 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the Fapi object.
-            description (bytes or str, optional): New description of the Fapi object. Defaults to None.
+            description (bytes or str): New description of the Fapi object. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -616,7 +726,7 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the Fapi object.
-            app_data (bytes or str, optional): Custom application data to be associated with the Fapi object. Defaults to None.
+            app_data (bytes or str): Custom application data to be associated with the Fapi object. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -658,7 +768,7 @@ class FAPI:
 
         Args:
             path (bytes or str): Path to the Fapi object.
-            certificate (bytes or str, optional): x509 certificate to be associated with the Fapi object. Defaults to None.
+            certificate (bytes or str): x509 certificate to be associated with the Fapi object. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -691,7 +801,7 @@ class FAPI:
         """Get the platform certificate and the so-called delta certificates.
 
         Args:
-            no_cert_ok (bool, optional): If True, an empty byte string is returned if no certificate is found. If False, in this case a TSS2_Exception is raised. Defaults to False.
+            no_cert_ok (bool): If True, an empty byte string is returned if no certificate is found. If False, in this case a TSS2_Exception is raised. Defaults to False.
 
         Raises:
             TSS2_Exception:  If Fapi returned an error code.
@@ -780,7 +890,7 @@ class FAPI:
             TSS2_Exception: If Fapi returned an error code.
 
         Returns:
-            Tuple[bytes, Any]: A tuple of the binary blob and its type (:const:`FAPI_ESYSBLOB.CONTEXTLOAD` or :const:`FAPI_ESYSBLOB.DESERIALIZE)`
+            Tuple[bytes, Any]: A tuple of the binary blob and its type (:const:`constants.FAPI_ESYSBLOB.CONTEXTLOAD` or :const:`constants.FAPI_ESYSBLOB.DESERIALIZE`)
         """
         path = _to_bytes_or_null(path)
         type_ = ffi.new("uint8_t *")
@@ -819,7 +929,7 @@ class FAPI:
         Args:
             policy_path (bytes or str): Path to the underlying policy.
             key_path (bytes or str): Path to the key associated with the policy Authorize.
-            policy_ref (bytes or str, optional): Additional application data (e.g. a reference to another policy). Defaults to None.
+            policy_ref (bytes or str): Additional application data (e.g. a reference to another policy). Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -874,7 +984,7 @@ class FAPI:
         Args:
             index (int): Index of the PCR (in the range of 0-23 in most cases).
             data (bytes or str): Input data to the extend operation.
-            log (bytes or str, optional): JSON-encoded event log data. Defaults to None.
+            log (bytes or str): JSON-encoded event log data. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -900,8 +1010,8 @@ class FAPI:
         Args:
             path (bytes or str): Path to the key used for signing.
             pcrs (List[int]): List of PCR indices to be included in the quote.
-            quote_type (bytes or str, optional): Type of quote to create. The default "TPM-Quote" is used if None is given. Defaults to None.
-            qualifying_data (bytes or str, optional): Additional application-defined data. Defaults to None.
+            quote_type (bytes or str): Type of quote to create. The default "TPM-Quote" is used if None is given. Defaults to None.
+            qualifying_data (bytes or str): Additional application-defined data. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -965,9 +1075,9 @@ class FAPI:
         Args:
             path (bytes or str): Path to the key used for verifying the signature.
             signature (bytes): Signature to the quote.
-            quote_info (bytes or str, optional): Quote info structure.
-            qualifying_data (bytes or str, optional): Additional application-defined data. Defaults to None.
-            pcr_log (bytes or str, optional): JSON-encoded PCR log entry.
+            quote_info (bytes or str): Quote info structure.
+            qualifying_data (bytes or str): Additional application-defined data. Defaults to None.
+            pcr_log (bytes or str): JSON-encoded PCR log entry.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -1006,9 +1116,9 @@ class FAPI:
         Args:
             path (bytes or str): Path to the NV storage area.
             size (int): Size of the storage area in bytes.
-            type_ (bytes or str, optional): Type of the storage area. A combination of `bitfield`, `counter`, `pcr`, `system`, `noda`. Defaults to None.
-            policy_path (bytes or str, optional): The path to the policy which will be associated with the storage area. Defaults to None.
-            auth_value (bytes or str, optional): Password to protect the new storage area. Defaults to None.
+            type_ (bytes or str): Type of the storage area. A combination of `bitfield`, `counter`, `pcr`, `system`, `noda`. Defaults to None.
+            policy_path (bytes or str): The path to the policy which will be associated with the storage area. Defaults to None.
+            auth_value (bytes or str): Password to protect the new storage area. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -1071,7 +1181,7 @@ class FAPI:
         Args:
             path (bytes or str): Path to the NV storage area.
             data (bytes or str): Input data to the extend operation.
-            log (bytes or str, optional): JSON-encoded event to be associated with this change. Defaults to None.
+            log (bytes or str): JSON-encoded event to be associated with this change. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
@@ -1126,162 +1236,97 @@ class FAPI:
         ret = lib.Fapi_WriteAuthorizeNv(self._ctx, nv_path, policy_path)
         _chkrc(ret)
 
-    def _register_callback(
-        self,
-        callback_type: CallbackType,
-        callback_wrapper: Callable,
-        unlock: bool = False,
-    ) -> Callable:
-        """(Un)register a C callback and tie it to a python wrapper. Does not call Fapi API calls.
-
-        Args:
-            callback_type (CallbackType): Type of callback. Each type can have up to one callback.
-            callback_wrapper (Callable): Python wrapper which is called by the C callback and will call a user-defined function.
-            unlock (bool, optional): True if the callback is to be freed again. Defaults to False.
-
-        Returns:
-            str: CFFI binding function
-        """
-        if unlock and self.callbacks[callback_type] is not None:
-            # unlock c callback
-            unlock_callback(CallbackType.FAPI_AUTH, self.callbacks[callback_type].name)  # type: ignore
-            self.callbacks[callback_type] = None
-
-            c_callback = ffi.NULL
-        else:
-            if (
-                callback_type not in self.callbacks
-                or self.callbacks[callback_type] is None
-            ):
-                # get c callback and lock it
-                self.callbacks[callback_type] = get_callback(callback_type)
-
-            # link callback wrapper to c function
-            callback_wrapper.__name__ = self.callbacks[callback_type].name  # type: ignore
-            ffi.def_extern()(callback_wrapper)
-
-            c_callback = self.callbacks[callback_type].c_function  # type: ignore
-
-        return c_callback
-
     def set_auth_callback(
         self,
-        callback: Optional[Callable[[str, str, Optional[bytes]], bytes]] = None,
-        user_data: Optional[Union[bytes, str]] = None,
+        callback: Optional[
+            Callable[[str, str, Optional[Any]], Union[bytes, str]]
+        ] = None,
+        user_data: Optional[Any] = None,
     ) -> None:
         """Register a callback that provides the password for Fapi objects when
         needed. Typically, this callback implements a password prompt. If `callback` is None, the callback function is reset.
 
         Args:
-            callback (Callable[[str, str, Optional[bytes]], bytes], optional): A callback function `callback(path, description, user_data=None)` which returns the password (:class:`bytes`). Defaults to None.
-            user_data (byte, optional): Bytes that will be handed to the callback. Defaults to None.
+            callback (Optional[Callable[[str, str, Optional[Any]], Union[bytes, str]]]): A callback function `callback(path, description, user_data=None)` which returns the password (:class:`str`). Defaults to None.
+            user_data (Any): Bytes that will be handed to the callback. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
+            RuntimeError: If callback is None and user_data is NOT None.
         """
+
         if callback is None and user_data is not None:
             raise RuntimeError("If callback is None, user_data must be None, too.")
 
-        if user_data is None:
-            user_data_len = 0
+        if callback:
+            cb_udata = _FAPI_CB_UDATA(callback, user_data)
+            cb_user_data_handle = ffi.new_handle(cb_udata)
+            # keep this alive, overwrite old so it be gc'd
+            self._callback_metadata["auth"] = cb_user_data_handle
+            lib_callback = lib._fapi_auth_callback
         else:
-            user_data_len = len(user_data)
-        user_data = _to_bytes_or_null(user_data)
+            lib_callback = ffi.NULL
+            cb_user_data_handle = ffi.NULL
+            self._callback_metadata["auth"] = None
 
-        def callback_wrapper(path, description, auth, user_data):
-            path = ffi.string(path).decode()
-            description = ffi.string(description).decode()
-            if user_data == ffi.NULL:
-                user_data = None
-            else:
-                user_data = bytes(
-                    ffi.unpack(ffi.cast("uint8_t *", user_data), user_data_len)
-                )
-            try:
-                auth_value = callback(path, description, user_data)
-            except Exception:
-                return lib.TSS2_FAPI_RC_CB_FAILURE
-
-            # auth value is cleaned up by the FAPI
-            auth[0] = ffi_malloc("char[]", auth_value)
-            return lib.TPM2_RC_SUCCESS
-
-        c_callback = self._register_callback(
-            CallbackType.FAPI_AUTH, callback_wrapper, unlock=callback is None
-        )
-
-        ret = lib.Fapi_SetAuthCB(self._ctx, c_callback, user_data)
+        ret = lib.Fapi_SetAuthCB(self._ctx, lib_callback, cb_user_data_handle)
         _chkrc(ret)
 
     def set_branch_callback(
         self,
-        callback: Optional[
-            Callable[[str, str, List[str], Optional[bytes]], int]
-        ] = None,
-        user_data: Optional[Union[bytes, str]] = None,
+        callback: Optional[Callable[[str, str, List[str], Optional[Any]], int]] = None,
+        user_data: Optional[Any] = None,
     ):
         """Set the Fapi policy branch callback, called to decide which policy path to take in a policy Or. If `callback` is None, the callback function is reset.
 
         Args:
-            callback (Callable[[str, str, List[str], Optional[bytes]], int], optional): A callback function `callback(path, description, branch_names, user_data=None)` which returns the index (:class:`int`) of the selected branch in `branch_names`. Defaults to None.
-            user_data (bytes or str, optional): Custom data passed to the callback function. Defaults to None.
+            callback (Callable[[str, str, List[str], Optional[bytes]], int]): A callback function `callback(path, description, branch_names, user_data=None)` which returns the index (:class:`int`) of the selected branch in `branch_names`. Defaults to None.
+            user_data (bytes or str): Custom data passed to the callback function. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
+            RuntimeError: If callback is None and user_data is NOT None.
         """
 
+        if callback is None:
+            _check_bug_fixed(
+                fixed_in="3.1",
+                backports=["2.4.3", "3.0.1", "3.1.0"],
+                details="A NULL FAPI SetBranchCallback Action might lead to crashes. See https://github.com/tpm2-software/tpm2-tss/pull/2500",
+            )
+
         if callback is None and user_data is not None:
-            raise ValueError("If callback is None, user_data must be None, too.")
+            raise RuntimeError("If callback is None, user_data must be None, too.")
 
-        if user_data is None:
-            user_data_len = 0
+        if callback:
+            cb_udata = _FAPI_CB_UDATA(callback, user_data)
+            cb_user_data_handle = ffi.new_handle(cb_udata)
+            # keep this alive, overwrite old so it be gc'd
+            self._callback_metadata["branch"] = cb_user_data_handle
+            lib_callback = lib._fapi_branch_callback
         else:
-            user_data_len = len(user_data)
-        user_data = _to_bytes_or_null(user_data)
-
-        def callback_wrapper(
-            path, description, branch_names, num_branches, selected_branch, user_data
-        ):
-            path = ffi.string(path).decode()
-            description = ffi.string(description).decode()
-            branch_names = [
-                ffi.string(branch_names[i]).decode() for i in range(0, num_branches)
-            ]
-            if user_data == ffi.NULL:
-                user_data = None
-            else:
-                user_data = bytes(
-                    ffi.unpack(ffi.cast("uint8_t *", user_data), user_data_len)
-                )
-            try:
-                selected_branch[0] = callback(
-                    path, description, branch_names, user_data
-                )
-            except Exception:
-                return lib.TSS2_FAPI_RC_GENERAL_FAILURE
-            return lib.TPM2_RC_SUCCESS
-
-        c_callback = self._register_callback(
-            CallbackType.FAPI_BRANCH, callback_wrapper, unlock=callback is None
-        )
-        ret = lib.Fapi_SetBranchCB(self._ctx, c_callback, user_data)
+            lib_callback = ffi.NULL
+            cb_user_data_handle = ffi.NULL
+            self._callback_metadata["branch"] = None
+        ret = lib.Fapi_SetBranchCB(self._ctx, lib_callback, cb_user_data_handle)
         _chkrc(ret)
 
     def set_sign_callback(
         self,
         callback: Optional[
-            Callable[[str, str, str, str, int, bytes, Optional[bytes]], bytes]
+            Callable[[str, str, str, str, int, bytes, Optional[Any]], bytes]
         ] = None,
-        user_data: Optional[Union[bytes, str]] = None,
+        user_data: Optional[Any] = None,
     ):
         """Set the Fapi signing callback which is called to satisfy the policy Signed. If `callback` is None, the callback function is reset.
 
         Args:
-            callback (Callable[[str, str, str, str, int, bytes, Optional[bytes]], bytes], optional): A callback function `callback(path, description, public_key, public_key_hint, hash_alg, data_to_sign, user_data=None)` which returns a signature (:class:`bytes`) of `data_to_sign`. Defaults to None.
-            user_data (bytes or str, optional): Custom data passed to the callback function. Defaults to None.
+            callback (Callable[[str, str, str, str, int, bytes, Optional[bytes]], bytes]): A callback function `callback(path, description, public_key, public_key_hint, hash_alg, data_to_sign, user_data=None)` which returns a signature (:class:`bytes`) of `data_to_sign`. Defaults to None.
+            user_data (Any): Custom data passed to the callback function. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
+            RuntimeError: If callback is None and user_data is NOT None.
         """
         _check_bug_fixed(
             fixed_in="3.2",
@@ -1291,75 +1336,35 @@ class FAPI:
         if callback is None and user_data is not None:
             raise RuntimeError("If callback is None, user_data must be None, too.")
 
-        if user_data is None:
-            user_data_len = 0
+        if callback:
+            cb_udata = _FAPI_CB_UDATA(callback, user_data)
+            cb_user_data_handle = ffi.new_handle(cb_udata)
+            # keep this alive, overwrite old so it be gc'd
+            self._callback_metadata["sign"] = cb_user_data_handle
+            lib_callback = lib._fapi_sign_callback
         else:
-            user_data_len = len(user_data)
-        user_data = _to_bytes_or_null(user_data)
+            lib_callback = ffi.NULL
+            cb_user_data_handle = ffi.NULL
+            self._callback_metadata["sign"] = None
 
-        def callback_wrapper(
-            path,
-            description,
-            public_key,
-            public_key_hint,
-            hash_alg,
-            data_to_sign,
-            data_to_sign_len,
-            signature,
-            signature_len,
-            user_data,
-        ):
-            path = ffi.string(path).decode()
-            description = ffi.string(description).decode()
-            public_key = ffi.string(public_key).decode()
-            public_key_hint = ffi.string(public_key_hint).decode()
-            data_to_sign = bytes(ffi.unpack(data_to_sign, data_to_sign_len))
-            if user_data == ffi.NULL:
-                user_data = None
-            else:
-                user_data = bytes(
-                    ffi.unpack(ffi.cast("uint8_t *", user_data), user_data_len,)
-                )
-            try:
-                signature_value = callback(
-                    path,
-                    description,
-                    public_key,
-                    public_key_hint,
-                    hash_alg,
-                    data_to_sign,
-                    user_data,
-                )
-            except Exception:
-                return lib.TSS2_FAPI_RC_CB_FAILURE
-
-            # signature is cleaned up by the FAPI
-            signature[0] = ffi_malloc("char[]", signature_value)
-            signature_len[0] = len(signature_value)
-            return lib.TPM2_RC_SUCCESS
-
-        c_callback = self._register_callback(
-            CallbackType.FAPI_SIGN, callback_wrapper, unlock=callback is None
-        )
-        ret = lib.Fapi_SetSignCB(self._ctx, c_callback, user_data)
+        ret = lib.Fapi_SetSignCB(self._ctx, lib_callback, cb_user_data_handle)
         _chkrc(ret)
 
     def set_policy_action_callback(
         self,
         callback: Optional[Callable[[str, str, Optional[bytes]], None]] = None,
-        user_data: Optional[Union[bytes, str]] = None,
+        user_data: Optional[Any] = None,
     ):
         """Set the policy Action callback which is called to satisfy the policy Action. If `callback` is None, the callback function is reset.
 
         Args:
-            callback (Callable[[str, str, Optional[bytes]], None], optional): A callback function `callback(path, action, user_data=None)`. Defaults to None.
-            user_data (bytes or str, optional): Custom data passed to the callback function. Defaults to None.
+            callback (Callable[[str, str, Optional[bytes]], None]): A callback function `callback(path, action, user_data=None)`. Defaults to None.
+            user_data (Any): Custom data passed to the callback function. Defaults to None.
 
         Raises:
             TSS2_Exception: If Fapi returned an error code.
+            RuntimeError: If callback is None and user_data is NOT None.
         """
-        if callback is None and user_data is not None:
-            raise ValueError("If callback is None, user_data must be None, too.")
 
         _check_bug_fixed(
             fixed_in="3.2",
@@ -1367,29 +1372,18 @@ class FAPI:
             details="FAPI Policy Action might lead to crashes. See https://github.com/tpm2-software/tpm2-tss/issues/2089",
         )
 
-        if user_data is None:
-            user_data_len = 0
+        if callback is None and user_data is not None:
+            raise RuntimeError("If callback is None, user_data must be None, too.")
+
+        if callback:
+            cb_udata = _FAPI_CB_UDATA(callback, user_data)
+            cb_user_data_handle = ffi.new_handle(cb_udata)
+            # keep this alive, overwrite old so it be gc'd
+            self._callback_metadata["policy"] = cb_user_data_handle
+            lib_callback = lib._fapi_policy_action_callback
         else:
-            user_data_len = len(user_data)
-        user_data = _to_bytes_or_null(user_data)
-
-        def callback_wrapper(path, action, user_data):
-            path = ffi.string(path).decode()
-            action = ffi.string(action).decode()
-            if user_data == ffi.NULL:
-                user_data = None
-            else:
-                user_data = bytes(
-                    ffi.unpack(ffi.cast("uint8_t *", user_data), user_data_len)
-                )
-            try:
-                callback(path, action, user_data)
-            except Exception:
-                return lib.TSS2_FAPI_RC_GENERAL_FAILURE
-            return lib.TPM2_RC_SUCCESS
-
-        c_callback = self._register_callback(
-            CallbackType.FAPI_POLICYACTION, callback_wrapper, unlock=callback is None
-        )
-        ret = lib.Fapi_SetPolicyActionCB(self._ctx, c_callback, user_data)
+            lib_callback = ffi.NULL
+            cb_user_data_handle = ffi.NULL
+            self._callback_metadata["policy"] = None
+        ret = lib.Fapi_SetPolicyActionCB(self._ctx, lib_callback, cb_user_data_handle)
         _chkrc(ret)
