@@ -7,8 +7,7 @@
 Along with helpers to go from string values to constants and constant values to string values.
 """
 from ._libtpm2_pytss import lib, ffi
-from tpm2_pytss.internal.utils import _CLASS_INT_ATTRS_from_string
-import pkgconfig
+from tpm2_pytss.internal.utils import _CLASS_INT_ATTRS_from_string, _lib_version_atleast
 
 
 class TPM_FRIENDLY_INT(int):
@@ -201,12 +200,33 @@ class TPM_FRIENDLY_INT(int):
         return self.__class__(int(self).__xor__(value))
 
     @staticmethod
-    def _fix_const_type(cls):
-        for k, v in vars(cls).items():
-            if not isinstance(v, int) or k.startswith("_"):
+    def _copy_and_set(dstcls, srccls):
+        """Copy class variables from srccls to dstcls
+
+        srccls must be a subclass on TPM_FRIENDLY_INT.
+        Sets the the variable type to dstcls when copying.
+        dstcls and srccls can be the same.
+        Skip setting destination if the dstcls variable already have the correct type.
+        """
+        if not issubclass(srccls, TPM_FRIENDLY_INT):
+            return
+        for k, v in vars(srccls).items():
+            dv = getattr(dstcls, k, None)
+            if not isinstance(v, int) or k.startswith("_") or type(dv) == type(dstcls):
                 continue
-            fv = cls(v)
-            setattr(cls, k, fv)
+            fv = dstcls(v)
+            setattr(dstcls, k, fv)
+
+    @staticmethod
+    def _fix_const_type(cls):
+        """Ensure constants in a TPM2 constant class have the correct type
+
+        We also copy constants from a superclass in case it's of the correct type.
+        """
+        for sc in cls.__mro__:
+            if sc in (TPM_FRIENDLY_INT, TPMA_FRIENDLY_INTLIST):
+                break
+            TPM_FRIENDLY_INT._copy_and_set(cls, sc)
         return cls
 
 
@@ -442,6 +462,7 @@ class TPM2_RH(TPM_FRIENDLY_INT):
 class TPM2_ALG(TPM_FRIENDLY_INT):
     ERROR = lib.TPM2_ALG_ERROR
     RSA = lib.TPM2_ALG_RSA
+    TDES = lib.TPM2_ALG_TDES
     SHA = lib.TPM2_ALG_SHA
     SHA1 = lib.TPM2_ALG_SHA1
     HMAC = lib.TPM2_ALG_HMAC
@@ -483,7 +504,9 @@ class TPM2_ALG(TPM_FRIENDLY_INT):
     LAST = lib.TPM2_ALG_LAST
 
 
-TPM2_ALG_ID = TPM2_ALG
+@TPM_FRIENDLY_INT._fix_const_type
+class TPM2_ALG_ID(TPM2_ALG):
+    pass
 
 
 @TPM_FRIENDLY_INT._fix_const_type
@@ -504,10 +527,13 @@ class TPM2_ECC(TPM_FRIENDLY_INT):
         "256": "NIST_P256",
         "384": "NIST_P384",
         "521": "NIST_P521",
+        "SM2": "SM2_P256",
     }
 
 
-TPM2_ECC_CURVE = TPM2_ECC
+@TPM_FRIENDLY_INT._fix_const_type
+class TPM2_ECC_CURVE(TPM2_ECC):
+    pass
 
 
 @TPM_FRIENDLY_INT._fix_const_type
@@ -642,7 +668,7 @@ class TPM2_SPEC(TPM_FRIENDLY_INT):
 
 
 @TPM_FRIENDLY_INT._fix_const_type
-class TPM2_GENERATED_VALUE(TPM_FRIENDLY_INT):
+class TPM2_GENERATED(TPM_FRIENDLY_INT):
     VALUE = lib.TPM2_GENERATED_VALUE
 
 
@@ -841,7 +867,7 @@ class TSS2_RC(TPM_BASE_RC):
     BASE_RC_BAD_KEY = lib.TSS2_BASE_RC_BAD_KEY
     BASE_RC_NO_HANDLE = lib.TSS2_BASE_RC_NO_HANDLE
 
-    if pkgconfig.installed("tss2-esapi", ">=3.0.0"):
+    if _lib_version_atleast("tss2-esapi", "3.0.0"):
         BASE_RC_NOT_PROVISIONED = lib.TSS2_BASE_RC_NOT_PROVISIONED
         BASE_RC_ALREADY_PROVISIONED = lib.TSS2_BASE_RC_ALREADY_PROVISIONED
 
@@ -906,7 +932,7 @@ class TSS2_RC(TPM_BASE_RC):
     ESYS_RC_MULTIPLE_DECRYPT_SESSIONS = lib.TSS2_ESYS_RC_MULTIPLE_DECRYPT_SESSIONS
     ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS = lib.TSS2_ESYS_RC_MULTIPLE_ENCRYPT_SESSIONS
     ESYS_RC_RSP_AUTH_FAILED = lib.TSS2_ESYS_RC_RSP_AUTH_FAILED
-    if pkgconfig.installed("tss2-fapi", ">=3.0.0"):
+    if _lib_version_atleast("tss2-fapi", "3.0.0"):
         FAPI_RC_GENERAL_FAILURE = lib.TSS2_FAPI_RC_GENERAL_FAILURE
         FAPI_RC_NOT_IMPLEMENTED = lib.TSS2_FAPI_RC_NOT_IMPLEMENTED
         FAPI_RC_BAD_REFERENCE = lib.TSS2_FAPI_RC_BAD_REFERENCE
@@ -946,6 +972,17 @@ class TSS2_RC(TPM_BASE_RC):
         FAPI_RC_NO_HANDLE = lib.TSS2_FAPI_RC_NO_HANDLE
         FAPI_RC_NOT_PROVISIONED = lib.TSS2_FAPI_RC_NOT_PROVISIONED
         FAPI_RC_ALREADY_PROVISIONED = lib.TSS2_FAPI_RC_ALREADY_PROVISIONED
+    if _lib_version_atleast("tss2-policy", "4.0.0"):
+        POLICY_RC_GENERAL_FAILURE = lib.TSS2_POLICY_RC_GENERAL_FAILURE
+        POLICY_RC_IO_ERROR = lib.TSS2_POLICY_RC_IO_ERROR
+        POLICY_RC_AUTHORIZATION_UNKNOWN = lib.TSS2_POLICY_RC_AUTHORIZATION_UNKNOWN
+        POLICY_RC_BAD_VALUE = lib.TSS2_POLICY_RC_BAD_VALUE
+        POLICY_RC_MEMORY = lib.TSS2_POLICY_RC_MEMORY
+        POLICY_RC_BAD_REFERENCE = lib.TSS2_POLICY_RC_BAD_REFERENCE
+        POLICY_RC_BAD_TEMPLATE = lib.TSS2_POLICY_RC_BAD_TEMPLATE
+        POLICY_RC_POLICY_NOT_CALCULATED = lib.TSS2_POLICY_RC_POLICY_NOT_CALCULATED
+        POLICY_RC_BUFFER_TOO_SMALL = lib.TSS2_POLICY_RC_BUFFER_TOO_SMALL
+        POLICY_RC_NULL_CALLBACK = lib.TSS2_POLICY_RC_NULL_CALLBACK
 
 
 @TPM_FRIENDLY_INT._fix_const_type
@@ -1014,6 +1051,20 @@ class TPM2_CAP(TPM_FRIENDLY_INT):
     VENDOR_PROPERTY = lib.TPM2_CAP_VENDOR_PROPERTY
 
 
+@TPM_FRIENDLY_INT._fix_const_type
+class TPMI_YES_NO(TPM_FRIENDLY_INT):
+    YES = lib.TPM2_YES
+    NO = lib.TPM2_NO
+
+
+@TPM_FRIENDLY_INT._fix_const_type
+class TPM_AT(TPM_FRIENDLY_INT):
+    ANY = lib.TPM_AT_ANY
+    ERROR = lib.TPM_AT_ERROR
+    PV1 = lib.TPM_AT_PV1
+    VEND = lib.TPM_AT_VEND
+
+
 @TPM_FRIENDLY_INT._fix_const_type
 class TPM2_PT(TPM_FRIENDLY_INT):
     NONE = lib.TPM2_PT_NONE
@@ -1070,11 +1121,13 @@ class TPM2_PT_FIRMWARE(TPM_FRIENDLY_INT):
 @TPM_FRIENDLY_INT._fix_const_type
 class TPM2_PT_HR(TPM_FRIENDLY_INT):
     LOADED_MIN = lib.TPM2_PT_HR_LOADED_MIN
+    NV_INDEX = lib.TPM2_PT_HR_NV_INDEX
     LOADED = lib.TPM2_PT_HR_LOADED
     LOADED_AVAIL = lib.TPM2_PT_HR_LOADED_AVAIL
     ACTIVE = lib.TPM2_PT_HR_ACTIVE
     ACTIVE_AVAIL = lib.TPM2_PT_HR_ACTIVE_AVAIL
     TRANSIENT_AVAIL = lib.TPM2_PT_HR_TRANSIENT_AVAIL
+    PERSISTENT = lib.TPM2_PT_HR_PERSISTENT
     PERSISTENT_AVAIL = lib.TPM2_PT_HR_PERSISTENT_AVAIL
     TRANSIENT_MIN = lib.TPM2_PT_HR_TRANSIENT_MIN
     PERSISTENT_MIN = lib.TPM2_PT_HR_PERSISTENT_MIN
@@ -1169,7 +1222,7 @@ class TPM2_HT(TPM_FRIENDLY_INT):
 
 
 @TPM_FRIENDLY_INT._fix_const_type
-class TPMA_SESSION(TPM_FRIENDLY_INT):
+class TPMA_SESSION(TPMA_FRIENDLY_INTLIST):
     CONTINUESESSION = lib.TPMA_SESSION_CONTINUESESSION
     AUDITEXCLUSIVE = lib.TPMA_SESSION_AUDITEXCLUSIVE
     AUDITRESET = lib.TPMA_SESSION_AUDITRESET
@@ -1179,7 +1232,7 @@ class TPMA_SESSION(TPM_FRIENDLY_INT):
 
 
 @TPM_FRIENDLY_INT._fix_const_type
-class TPMA_LOCALITY(TPM_FRIENDLY_INT):
+class TPMA_LOCALITY(TPMA_FRIENDLY_INTLIST):
     ZERO = lib.TPMA_LOCALITY_TPM2_LOC_ZERO
     ONE = lib.TPMA_LOCALITY_TPM2_LOC_ONE
     TWO = lib.TPMA_LOCALITY_TPM2_LOC_TWO
@@ -1195,6 +1248,44 @@ class TPMA_LOCALITY(TPM_FRIENDLY_INT):
             raise ValueError("Extended Localities must be less than 256")
         return x
 
+    @classmethod
+    def parse(cls, value: str) -> "TPMA_LOCALITY":
+        """Converts a string of | separated localities or an extended locality into a TPMA_LOCALITY instance
+
+        Args:
+            value (str): The string "bitwise" expression of the localities or the extended locality.
+
+        Returns:
+            The locality or set of localities as a TPMA_LOCALITY instance.
+
+        Raises:
+            TypeError: If the value is not a str.
+            ValueError: If a field portion of the str does not match a constant.
+
+        Examples:
+            TPMA_LOCALITY.parse("zero|one") -> 0x03
+            TPMA_LOCALITY.parse("0xf0") -> 0xf0
+        """
+        try:
+            return cls(value, base=0)
+        except ValueError:
+            pass
+        return super().parse(value)
+
+    def __str__(self) -> str:
+        """Given a set of localities or an extended locality, return the string representation
+
+        Returns:
+            (str): a bitwise string value of the localities or the exteded locality.
+
+        Example:
+            str(TPMA_LOCALITY.THREE|TPMA_LOCALITY.FOUR) -> 'three|four'
+            str(TPMA_LOCALITY(0xf0)) -> '0xf0'
+        """
+        if self > 31:
+            return f"{self:#x}"
+        return super().__str__()
+
 
 @TPM_FRIENDLY_INT._fix_const_type
 class TPM2_NT(TPM_FRIENDLY_INT):
@@ -1264,7 +1355,9 @@ class TPM2_CLOCK(TPM_FRIENDLY_INT):
     COARSE_FASTER = lib.TPM2_CLOCK_COARSE_FASTER
 
 
-TPM2_CLOCK_ADJUST = TPM2_CLOCK
+@TPM_FRIENDLY_INT._fix_const_type
+class TPM2_CLOCK_ADJUST(TPM2_CLOCK):
+    pass
 
 
 @TPM_FRIENDLY_INT._fix_const_type
@@ -1311,6 +1404,7 @@ class TPMA_CC(TPMA_FRIENDLY_INTLIST):
     )
     COMMANDINDEX_MASK = lib.TPMA_CC_COMMANDINDEX_MASK
     COMMANDINDEX_SHIFT = lib.TPMA_CC_COMMANDINDEX_SHIFT
+    RESERVED1_MASK = lib.TPMA_CC_RESERVED1_MASK
     NV = lib.TPMA_CC_NV
     EXTENSIVE = lib.TPMA_CC_EXTENSIVE
     FLUSHED = lib.TPMA_CC_FLUSHED
@@ -1376,27 +1470,31 @@ class TPMA_ALGORITHM(TPM_FRIENDLY_INT):
     SYMMETRIC = lib.TPMA_ALGORITHM_SYMMETRIC
     HASH = lib.TPMA_ALGORITHM_HASH
     OBJECT = lib.TPMA_ALGORITHM_OBJECT
+    RESERVED1_MASK = lib.TPMA_ALGORITHM_RESERVED1_MASK
     SIGNING = lib.TPMA_ALGORITHM_SIGNING
     ENCRYPTING = lib.TPMA_ALGORITHM_ENCRYPTING
     METHOD = lib.TPMA_ALGORITHM_METHOD
 
 
 @TPM_FRIENDLY_INT._fix_const_type
-class TPMA_PERMANENT(TPM_FRIENDLY_INT):
+class TPMA_PERMANENT(TPMA_FRIENDLY_INTLIST):
     OWNERAUTHSET = lib.TPMA_PERMANENT_OWNERAUTHSET
     ENDORSEMENTAUTHSET = lib.TPMA_PERMANENT_ENDORSEMENTAUTHSET
     LOCKOUTAUTHSET = lib.TPMA_PERMANENT_LOCKOUTAUTHSET
+    RESERVED1_MASK = lib.TPMA_PERMANENT_RESERVED1_MASK
     DISABLECLEAR = lib.TPMA_PERMANENT_DISABLECLEAR
     INLOCKOUT = lib.TPMA_PERMANENT_INLOCKOUT
     TPMGENERATEDEPS = lib.TPMA_PERMANENT_TPMGENERATEDEPS
+    RESERVED2_MASK = lib.TPMA_PERMANENT_RESERVED2_MASK
 
 
 @TPM_FRIENDLY_INT._fix_const_type
-class TPMA_STARTUP(TPM_FRIENDLY_INT):
+class TPMA_STARTUP(TPMA_FRIENDLY_INTLIST):
     CLEAR_PHENABLE = lib.TPMA_STARTUP_CLEAR_PHENABLE
     CLEAR_SHENABLE = lib.TPMA_STARTUP_CLEAR_SHENABLE
     CLEAR_EHENABLE = lib.TPMA_STARTUP_CLEAR_EHENABLE
     CLEAR_PHENABLENV = lib.TPMA_STARTUP_CLEAR_PHENABLENV
+    CLEAR_RESERVED1_MASK = lib.TPMA_STARTUP_CLEAR_RESERVED1_MASK
     CLEAR_ORDERLY = lib.TPMA_STARTUP_CLEAR_ORDERLY
 
 
@@ -1418,6 +1516,12 @@ class TPM2_MAX(TPM_FRIENDLY_INT):
     CONTEXT_SIZE = lib.TPM2_MAX_CONTEXT_SIZE
 
 
+@TPM_FRIENDLY_INT._fix_const_type
+class TPMA_MODES(TPMA_FRIENDLY_INTLIST):
+    FIPS_140_2 = lib.TPMA_MODES_FIPS_140_2
+    RESERVED1_MASK = lib.TPMA_MODES_RESERVED1_MASK
+
+
 #
 # We specifically keep these constants around even when FAPI is missing so they may be used
 # without conditional worry and we DONT use lib prefix here because the constants are only
@@ -1427,3 +1531,9 @@ class TPM2_MAX(TPM_FRIENDLY_INT):
 class FAPI_ESYSBLOB(TPM_FRIENDLY_INT):
     CONTEXTLOAD = 1
     DESERIALIZE = 2
+
+
+@TPM_FRIENDLY_INT._fix_const_type
+class TSS2_POLICY_PCR_SELECTOR(TPM_FRIENDLY_INT):
+    PCR_SELECT = 0
+    PCR_SELECTION = 1
